# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import QueryLanguageClass
from .....schema_classes import QueryPropertiesClass
from .....schema_classes import QuerySourceClass
from .....schema_classes import QueryStatementClass
from .....schema_classes import QuerySubjectClass
from .....schema_classes import QuerySubjectsClass


QueryLanguage = QueryLanguageClass
QueryProperties = QueryPropertiesClass
QuerySource = QuerySourceClass
QueryStatement = QueryStatementClass
QuerySubject = QuerySubjectClass
QuerySubjects = QuerySubjectsClass

# fmt: on
